import pandas as pd
from sklearn import preprocessing
from numpy import mean
from numpy import std
import datetime
from catboost import CatBoostClassifier, Pool
import lightgbm as lgb
from xgboost import XGBClassifier
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import RepeatedStratifiedKFold
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pyplot import plot,savefig
import warnings
warnings.filterwarnings('ignore')
np.random.seed(42)

def printlog(info):
    nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("\n" + "==========" * 8 + "%s" % nowtime)
    print(info + '...\n\n')
# ================================================================================

# ================================================================================
printlog("step: read data...")

# f1 = open(r'F:/paper/code/1over_sample/over_sample_train.csv')
# f2 = open(r'F:/paper/code/1over_sample/over_sample_test.csv')
# f1 = open(r'F:/paper/code/1under_sample/under_sample_train.csv')
# f2 = open(r'F:/paper/code/1under_sample/under_sample_train.csv')
f1 = open(r'F:/paper/code/1SMOTETomek_sample/SMOTETomek_sample_train.csv')
f2 = open(r'F:/paper/code/1SMOTETomek_sample/SMOTETomek_sample_test.csv')

train = pd.read_csv(f1)
test = pd.read_csv(f2)
train = np.array(train)
test = np.array(test)

x_train = train[:, :-1]
y_train = train[:, -1]
Y_train = y_train/10

x_test = test[:, :-1]
y_test = test[:, -1]
Y_test = y_test/10

min_max_scaler = preprocessing.MinMaxScaler()
x_train = min_max_scaler.fit_transform(x_train)
x_test = min_max_scaler.fit_transform(x_test)

y_train = Y_train
y_test = Y_test

# ================================================================================
# DNN+ENCODER+PCA
# ================================================================================
printlog("step: DNN+ENCODER+PCA...")

# ================================================================================
#                DNN
# ================================================================================
import h5py
from keras.models import Sequential,load_model,Model
from keras.layers import Input, Dense
dnn = Model()
inputs = Input((20,))
h1 = Dense(20, activation='sigmoid')(inputs)
h2 =Dense(100, activation='relu')(h1)
h3 = Dense(50, activation='relu')(h2)
h4 = Dense(20, activation='relu')(h3)
outputs = Dense(6, activation='softmax')(h4)
dnn = Model(inputs=inputs,outputs=outputs)
dnn.compile( optimizer='adam',loss='sparse_categorical_crossentropy', metrics=['accuracy'])
dnn_history = dnn.fit(x_train, y_train, epochs=1000, batch_size = 500,validation_data=(x_train, y_train))

dnn_save_path = './dnn_s.h5'
dnn.save(dnn_save_path)


load_dnn_model = load_model('dnn_s.h5')
load_dnn_model.layers.pop()

new_dnn_model = Model(inputs=inputs ,outputs=h4 )
new_dnn_model.compile( optimizer='adam',loss='sparse_categorical_crossentropy', metrics=['accuracy'])
new_dnn_history = new_dnn_model.fit(x_train, y_train, epochs=1000, batch_size = 500,validation_data=(x_train, y_train))
# new_dnn_model.summary()
new_dnn_model.save('new_dnn_s.h5')

x_1_train = new_dnn_model.predict(x_train)
x_1_test = new_dnn_model.predict(x_test)




# ================================================================================
#                  ENCODER
# ================================================================================
from keras.layers import Input, Dense
from keras.models import Model
from keras import regularizers
import pickle

input_dim = x_train.shape[1] # 20

encoding_dim = 20

input_layer = Input(shape=(input_dim, ))
encoder_layer_1 = Dense(50, activation="relu", activity_regularizer=regularizers.l1(10e-5))(input_layer)#relu防止过拟合
encoder_layer_2 = Dense(encoding_dim, activation="relu")(encoder_layer_1)#activation='sigmoid'

encoder = Model(inputs=input_layer, outputs=encoder_layer_2)
encoder.compile(optimizer='adam', loss='sparse_categorical_crossentropy', metrics=['accuracy'])
encoder_history = encoder.fit(x_train, y_train, epochs=1000,batch_size = 500, validation_data=(x_train, y_train))

x_2_train = encoder.predict(x_train)
x_2_test = encoder.predict(x_test)


encoder_save_path = './encoder.h5'
encoder.save(encoder_save_path)

# ================================================================================
#                       PCA
# ================================================================================
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
import joblib

pca1 = PCA(n_components=10)
X_1 = pca1.fit_transform(x_train)
X_2 = pca1.fit_transform(x_test)
# print (X_1.shape)#(1655, 10)
x_3_train = X_1
x_3_test = X_2

# ================================================================================

# ================================================================================
printlog("step: A single learner...")


clf_1 = RandomForestClassifier(n_estimators=20,
                                max_depth= 10)
clf_2 = XGBClassifier(    max_depth=10,
                          learning_rate=0.1,
                          n_estimators=20,
                          n_jobs=-1,
                          colsample_bytree=0.8,
                          subsample=0.8)
clf_3 = CatBoostClassifier(learning_rate=0.1, depth=10,
                              n_estimators=20,verbose=False)
clf_4 = lgb.LGBMClassifier(boosting_type='gbdt',objective='binary',
        learning_rate=0.1, n_estimators=20, num_leaves=10, silent=True,max_depth=10,
        subsample=0.8,colsample_bytree=0.8)
clf_5 = GradientBoostingClassifier(learning_rate=0.1, subsample=0.8, max_depth=10, n_estimators=20)# gdbt

clfs = [clf_1,clf_2,clf_4,clf_5]#clf_1,clf_2,clf_3,clf_4
data_train = np.zeros((x_train.shape[0], len(clfs)))
data_test = np.zeros((x_test.shape[0], len(clfs)))

# ================================================================================

# ================================================================================
printlog("step: stacking...")

Y_train = Y_train.astype("str")
Y_test  = Y_test.astype("str")
n_splits = 15
# skf = StratifiedKFold(n_splits=5, shuffle=False)
from sklearn.model_selection import RepeatedStratifiedKFold
skf = RepeatedStratifiedKFold(n_splits=5, n_repeats=3, random_state=1)


for i, clf in enumerate(clfs):
    # print(i,clf)

    d_test = np.zeros((x_test.shape[0], n_splits))
    for fold, (train_idx, val_idx) in enumerate(skf.split(x_train, Y_train)):

        train_x = x_train[train_idx]#(92164, 20)
        # print(train_x.shape)
        train_y = Y_train[train_idx]#(92164,)
        # print(train_y.shape)
        val_x =x_train[val_idx]#(23041, 20)
        # print(val_x.shape)
        val_y = Y_train[val_idx]#(23041,)
        # print(val_y.shape)

        clf.fit(train_x, train_y)#(92164, 20),(92164,)
        # clf.fit(train_pool)#clf_3
        pred_train = clf.predict(train_x)
        pred_val = clf.predict(val_x)
        data_train[val_idx, i] = pred_val
        # data_train[val_idx] =pred_val#clf_3
        # a = data_train[val_idx, i]#(23041,)
        d_test[:, fold] = clf.predict(x_test)
        # d_test = clf.predict(x_test)#clf_3
        # b = d_test[:, fold]#(28817, )
        score_train  = cross_val_score(clf,train_x,pred_train,scoring='accuracy',error_score='raise')
    mse = mean(score_train)
    print (str(i)+'clf_train Accuracy：%.4f ', mse)

    data_test[:, i] = d_test.mean(axis=1)
# x_1_train = np.hstack([x_train ,x_1_train])
x_1_train = np.hstack([x_1_train,data_train])
# x_1_test = np.hstack([x_test ,x_1_test])
x_1_test = np.hstack([x_1_test,data_test])

# x_2_train = np.hstack([x_train ,x_2_train])
x_2_train = np.hstack([x_2_train,data_train])
# x_2_test = np.hstack([x_test ,x_2_test])
x_2_test  = np.hstack([x_2_test,data_test])

# x_3_train = np.hstack([x_train ,x_3_train])
x_3_train = np.hstack([x_3_train,data_train])
# x_3_test = np.hstack([x_test ,x_3_test])
x_3_test  = np.hstack([x_3_test,data_test])

data_train_1 = pd.DataFrame(x_1_train)
data_test_1 = pd.DataFrame(x_1_test)

data_train_2 = pd.DataFrame(x_2_train)
data_test_2 = pd.DataFrame(x_2_test)

data_train_3 = pd.DataFrame(x_3_train)
data_test_3 = pd.DataFrame(x_3_test)

y_train = Y_train.astype("str")
y_test  = Y_test.astype("str")


# clf = lgb.LGBMClassifier(boosting_type='gbdt',objective='binary',
#         learning_rate=0.1, n_estimators=20, num_leaves=10, silent=True,max_depth=10,
#         subsample=0.8,colsample_bytree=0.8)
clf = GradientBoostingClassifier(learning_rate=0.1, subsample=0.8, max_depth=10, n_estimators=20)# gdbt
cv = RepeatedStratifiedKFold(n_splits=5, n_repeats=3, random_state=1)

clf.fit(x_train, y_train)
# clf.fit(data_test, y_test)
pre_ = clf.predict (x_train)
scores = cross_val_score(clf,x_train, pre_,  cv = cv, scoring='accuracy' ,n_jobs=-1,error_score='raise')
loss_ = cross_val_score(clf, x_train, pre_,  cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')
pre_t_ = clf.predict (x_test)
scores_t = cross_val_score(clf,x_test, pre_t_, cv = cv, scoring='accuracy' ,n_jobs=-1,error_score='raise')
loss_t_ = cross_val_score(clf, x_test, pre_t_,  cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')

# Stacking-GDBT
clf.fit(data_train, y_train)
# clf.fit(data_test, y_test)
pre_y = clf.predict (data_train)
n_scores = cross_val_score(clf,x_train, pre_y, cv = cv, scoring='accuracy' ,n_jobs=-1,error_score='raise')
loss = cross_val_score(clf, x_train, pre_y, cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')
pre_y_t = clf.predict (data_test)
n_scores_t = cross_val_score(clf,x_test, pre_y_t, cv = cv, scoring='accuracy' ,n_jobs=-1,error_score='raise')
loss_t = cross_val_score(clf, x_test, pre_y_t, cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')






# # dnn+stacking
# clf.fit(data_train_1, y_train)
# pre_1 = clf.predict (data_train_1)
# n_scores_1 = cross_val_score(clf,x_train, pre_1, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
# loss_1 = cross_val_score(clf,x_train, pre_1, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')
#
# # clf.fit(data_test_1,y_test)
# pre_1_test = clf.predict (data_test_1)
# n_scores_test_1 = cross_val_score(clf,x_test, pre_1_test, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
# loss_test_1 = cross_val_score(clf,x_test, pre_1_test, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')


# encoder+stacking
clf.fit(data_train_2, y_train)
pre_2 = clf.predict (data_train_2)
n_scores_2 = cross_val_score(clf,x_train, pre_2, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
loss_2 = cross_val_score(clf,x_train, pre_2, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')

# clf.fit(data_test_2,y_test)
pre_2_test = clf.predict (data_test_2)
n_scores_test_2 = cross_val_score(clf,x_test, pre_2_test, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
loss_test_2 = cross_val_score(clf,x_test, pre_2_test, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')


# pca+stacking
clf.fit(data_train_3, y_train)
pre_3 = clf.predict (data_train_3)
n_scores_3 = cross_val_score(clf,x_train, pre_3, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
loss_3 = cross_val_score(clf,x_train, pre_3, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')


# clf.fit(data_test_3,y_test)
pre_3_test = clf.predict (data_test_3)
n_scores_test_3 = cross_val_score(clf,x_test, pre_3_test, cv = cv, scoring='accuracy',n_jobs=-1,error_score='raise')
loss_test_3 = cross_val_score(clf,x_test, pre_3_test, cv = cv, scoring='neg_log_loss',n_jobs=-1,error_score='raise')

plt.rcParams.update({'font.family': 'Times New Roman', 'font.size': 14.0})
from brokenaxes import brokenaxes
plt.rcParams['figure.figsize']=(14.0,10.0)
bax = brokenaxes(
                 xlims=[(0,21)],
                 ylims=((0.175, 0.501), (0.875, 1.01)),
                 hspace=0.25,
                 despine=False,
                 diag_color='r',
                 )
bax.plot(scores,
         linestyle='-',
         linewidth=2,
         color = 'green',
         markersize = 4,
         marker='s',
         label ='train_acc(GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(scores_t,
         linestyle='--',
         linewidth=2,
         color = 'green',
         markersize = 4,
         marker='s',
         label='test_acc(GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(n_scores,
         linestyle='-',
         linewidth=2,
         color = 'blue',
         markersize = 4,
         marker='^',
         label ='train_acc(Stacking(GDBT))',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(n_scores_t,
         linestyle='--',
         linewidth=2,
         color = 'blue',
         markersize = 4,
         marker='^',
         label='test_acc(Stacking(GDBT))',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(n_scores_2,
         linestyle='-',
         linewidth=2,
         color = 'brown',
         markersize = 4,
         marker='o',
         label ='train_acc(Encoder-FE+GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )

bax.plot(n_scores_test_2,
         linestyle='--',
         linewidth=2,
         color = 'brown',
         markersize = 4,
         marker='o',
         label='test_acc(Encoder-FE+GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )


bax.plot(-loss_,
         linestyle='-.',
         linewidth=2,
         color = 'red',
         markersize = 4,
         marker='p',
         label ='train_loss(GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )

bax.plot(-loss_t_,
         linestyle=':',
         linewidth=2,
         color = 'red',
         markersize = 4,
         marker='p',
         label='test_loss(GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(-loss,
         linestyle='-.',
         linewidth=2,
         color = 'orange',
         markersize = 4,
         marker='+',
         label ='train_loss(Stacking(GDBT))',
         markeredgecolor='black',
         markerfacecolor='black',
         )

bax.plot(-loss_t,
         linestyle=':',
         linewidth=2,
         color = 'orange',
         markersize = 4,
         marker='+',
         label='test_loss(Stacking(GDBT))',
         markeredgecolor='black',
         markerfacecolor='black',
         )
bax.plot(-loss_2,
         linestyle='-.',
         linewidth=2,
         color = 'darkorchid',
         markersize = 4,
         marker='>',
         label ='train_loss(Encoder-FE+GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )

bax.plot(-loss_test_2,
         linestyle=':',
         linewidth=2,
         color = 'darkorchid',
         markersize = 4,
         marker='>',
         label='test_loss(Encoder-FE+GDBT)',
         markeredgecolor='black',
         markerfacecolor='black',
         )


bax.set_ylabel('Accuracy and Loss', size=20, family='Times New Roman')
bax.set_xlabel('Number of Repeated Sampling', size=20, family='Times New Roman')
bax.legend(loc=1,prop={'family':'Times New Roman','size':14})
# plt.legend()
# savefig('test.jpg', dpi=1000)
# savefig('final_SMOTE.jpg')
# savefig('final_compare_over.jpg')
savefig('final_compare_SMOTE.jpg')
plt.clf()

