import pandas as pd
from sklearn import preprocessing
from numpy import mean
from numpy import std
import datetime
from catboost import CatBoostClassifier, Pool
import lightgbm as lgb
from xgboost import XGBClassifier
from sklearn.model_selection import cross_val_score
from sklearn.ensemble import RandomForestClassifier, GradientBoostingClassifier
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import RepeatedStratifiedKFold
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.pyplot import plot,savefig
import warnings
warnings.filterwarnings('ignore')
np.random.seed(42)

def printlog(info):
    nowtime = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')
    print("\n" + "==========" * 8 + "%s" % nowtime)
    print(info + '...\n\n')
# ================================================================================

# ================================================================================
printlog("step: read data...")

# f1 = open(r'F:/paper/code/1over_sample/over_sample_train.csv')
# f2 = open(r'F:/paper/code/1over_sample/over_sample_test.csv')
f1 = open(r'F:/paper/code/1under_sample/under_sample_train.csv')
f2 = open(r'F:/paper/code/1under_sample/under_sample_train.csv')
# f1 = open(r'F:/paper/code/1SMOTETomek_sample/SMOTETomek_sample_train.csv')
# f2 = open(r'F:/paper/code/1SMOTETomek_sample/SMOTETomek_sample_test.csv')

train = pd.read_csv(f1)
test = pd.read_csv(f2)
train = np.array(train)
test = np.array(test)

x_train = train[:, :-1]
y_train = train[:, -1]
Y_train = y_train/10

x_test = test[:, :-1]
y_test = test[:, -1]
Y_test = y_test/10

min_max_scaler = preprocessing.MinMaxScaler()
x_train = min_max_scaler.fit_transform(x_train)
x_test = min_max_scaler.fit_transform(x_test)

y_train = Y_train.astype("str")
y_test  = Y_test.astype("str")
# ================================================================================

# ================================================================================
printlog("step: stacking...")

import warnings
warnings.filterwarnings('ignore')
np.random.seed(42)


clf_1 = RandomForestClassifier(n_estimators=20,max_depth= 10)
clf_2 = XGBClassifier( max_depth=10, learning_rate=0.1,  n_estimators=20, colsample_bytree=0.8,subsample=0.8)
clf_3 = CatBoostClassifier(learning_rate=0.1, depth=10, n_estimators=20,verbose=False)
clf_4 = lgb.LGBMClassifier(boosting_type='gbdt',objective='binary',
        learning_rate=0.1, n_estimators=20, num_leaves=10, silent=True,max_depth=10,
        subsample=0.8,colsample_bytree=0.8)
clf_5 = GradientBoostingClassifier(learning_rate=0.1, subsample=0.8, max_depth=10, n_estimators=20)# gdbt

clfs = [clf_1,clf_2,clf_4,clf_5]#clf_1,clf_2,clf_3,clf_4
data_train = np.zeros((x_train.shape[0], len(clfs)))
data_test = np.zeros((x_test.shape[0], len(clfs)))

'''5折stacking'''

n_splits = 15
from sklearn.model_selection import RepeatedStratifiedKFold
skf = RepeatedStratifiedKFold(n_splits=5, n_repeats=3, random_state=1)
# for (train_idx, val_idx) in skf.split(x_train, y_train):
    # print((train_idx, val_idx))
    # print((train_idx.shape, val_idx.shape))#((92164,), (23041,))
    # print((len(train_idx), len(val_idx)))# (92164, 23041)
# catboost data_pool
# train_pool = Pool(x_train,y_train)
# test_pool = Pool(x_test,y_test)


for i, clf in enumerate(clfs):
    print(i,clf)

    d_test = np.zeros((x_test.shape[0], n_splits))
    for fold, (train_idx, val_idx) in enumerate(skf.split(x_train, y_train)):

        train_x = x_train[train_idx]#(92164, 20)
        # print(train_x.shape)
        train_y = y_train[train_idx]#(92164,)
        # print(train_y.shape)
        val_x =x_train[val_idx]#(23041, 20)
        # print(val_x.shape)
        val_y = y_train[val_idx]#(23041,)
        # print(val_y.shape)

        clf.fit(train_x, train_y)#(92164, 20),(92164,)
        # clf.fit(train_pool)#clf_3
        pred_train = clf.predict(train_x)

        pred_val = clf.predict(val_x)

        data_train[val_idx, i] = pred_val

        # a = data_train[val_idx, i]#(23041,)
        d_test[:, fold] = clf.predict(x_test)
        # d_test = clf.predict(x_test)#clf_3
        #d_test[:, fold]#(28817, )
        # score_train  = cross_val_score(clf,train_x,pred_train,scoring='accuracy',error_score='raise')
        # score_val = cross_val_score(clf,val_y, pred_val, scoring='accuracy',error_score='raise')


        data_test[:, i] = d_test.mean(axis=1)

data_train = pd.DataFrame(data_train)
data_test = pd.DataFrame(data_test)

# print (data_train.shape)
# print (data_test.shape)

clf = GradientBoostingClassifier(learning_rate=0.1, subsample=0.8, max_depth=10, n_estimators=20)# gdbt
# clf = lgb.LGBMClassifier(boosting_type='gbdt',objective='binary',
#         learning_rate=0.1, n_estimators=20, num_leaves=10, silent=True,max_depth=10,
#         subsample=0.8,colsample_bytree=0.8)
clf.fit(data_train, y_train)
# clf.fit(data_test, y_test)

cv = RepeatedStratifiedKFold(n_splits=5, n_repeats=3, random_state=1)
pre_y = clf.predict (data_train)
n_scores = cross_val_score(clf,x_train, pre_y, cv = cv, scoring='accuracy' ,n_jobs=-1,error_score='raise')
loss = cross_val_score(clf, x_train, pre_y, cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')



pre_y_t = clf.predict (data_test)
n_scores_t = cross_val_score(clf,x_test, pre_y_t, cv = cv, scoring='accuracy', n_jobs=-1,error_score='raise')
loss_t = cross_val_score(clf, x_test, pre_y_t, cv = cv, scoring='neg_log_loss', n_jobs=-1,error_score='raise')





# ================================================================================

# ================================================================================
printlog("step: saving model ...")
